"""
Validator type.

- type: option ["*": no limit, digit only for no variable list=None(no check), [""](no variable)]
    - int: (min, max)
    - float: (min, max, digit)
    - sympy_float: (digit)
    - sympy: (variable list)
    - ilist: (shape)
    - list: (shape, variable list, digit)
    - site: (use variable?)
    - bond: (use variable?)
    - site_bond: (use variable?)
    - vector_site_bond: (use variable?)
    - orbital_site_bond: (use variable?)
"""

import numpy as np
import sympy as sp
from qtdraw.util.util_str import str_to_numpy, apply_to_numpy


# ==================================================
def validator_int(text, opt):
    """
    Validator for int.

    Args:
        text (str): int string.
        opt (tuple, optional): option, (min, max).

    Returns:
        - (str) -- formatted string if it is valid, otherwise None.
    """
    r_min, r_max = opt

    # type conversion.
    try:
        s = int(text)
    except:
        return None

    # range check.
    if r_min != "*" and s < int(r_min):
        return None
    if r_max != "*" and s > int(r_max):
        return None

    return str(s)


# ==================================================
def validator_float(text, opt):
    """
    Validator for float.

    Args:
        text (str): int string.
        opt (tuple, optional): option, (min, max, digit).

    Returns:
        - (str) -- formatted string if it is valid, otherwise None.
    """
    r_min, r_max, digit = opt

    # type conversion.
    try:
        s = float(text)
    except:
        return None

    # range check.
    if r_min != "*" and s < float(r_min):
        return None
    if r_max != "*" and s > float(r_max):
        return None
    s = np.round(s, digit)
    s = format(s, f".0{digit}f")

    return s


# ==================================================
def validator_sympy_float(text, opt):
    """
    Validator for sympy.

    Args:
        text (str): sympy string.
        opt (int, optional): option, digit.

    Returns:
        - (str) -- formatted string if it is valid, otherwise None.
    """
    digit = opt

    # convert to sympy, (1x1).
    try:
        s = str_to_numpy(text, digit, None, ())
    except:
        return None
    if s is None:
        return None
    s = s.tolist()
    if digit is not None:
        s = format(s, f".0{digit}f")
    else:
        s = str(s)

    return s


# ==================================================
def validator_sympy(text, opt):
    """
    Validator for sympy.

    Args:
        text (str): sympy string.
        opt (list, optional): option, var.

    Returns:
        - (str) -- LaTeX string if it is valid, otherwise None.
    """
    var = opt

    # convert to sympy, (1x1).
    try:
        s = str_to_numpy(text, None, var, ())
    except:
        return None
    if s is None:
        return None
    s = s.tolist()
    s = sp.latex(s)

    return s


# ==================================================
def validator_ilist(s, opt):
    """
    Validator for int list.

    Args:
        s (str): int list string.
        opt (tuple, optional): option, shape, (), (n,), (n,m), ...

    Returns:
        - (str) -- formatted string if it is valid, otherwise None.
    """
    shape = opt

    # convert to sympy with shape and no varialbes.
    try:
        s = str_to_numpy(s, check_var=[""], check_shape=shape)
    except:
        return None
    if s is None:
        return None

    # type conversion.
    try:
        s = s.astype(int)
    except:
        return None

    s = str(s.tolist())

    return s


# ==================================================
def validator_list(s, opt):
    """
    Validator for sympy list.

    Args:
        s (str): sympy list string.
        opt (tuple, optional): option, (shape, var, digit).

    Returns:
        - (str) -- formatted string if it is valid, otherwise None.
    """
    shape, var, digit = opt

    # convert to sympy with shape and no varialbes.
    try:
        s = str_to_numpy(s, digit, var, check_shape=shape)
    except:
        return None
    if s is None:
        return None
    if digit is not None and var is not None and len("".join(var)) == 0:
        s = apply_to_numpy(lambda x: format(x, f".0{digit}f"), s)

    s = str(s.tolist()).replace("'", "")

    return s


# ==================================================
def validator_site(s, use_var=False):
    """
    Validator for site.

    Args:
        s (str): site string, [x,y,z].
        use_var (bool, optional): use [x,y,z] for site/bond ?

    Returns:
        - (str) -- input s if it is valid, otherwise None.
    """
    if use_var:
        var = ["x", "y", "z"]
    else:
        var = [""]
    try:
        status = str_to_numpy(s, None, var, (3,))
    except:
        return None
    if status is not None:
        return s
    else:
        return None


# ==================================================
def validator_bond(s, use_var=False):
    """
    Validator for bond.

    Args:
        s (str): bond string.
        use_var (bool, optional): use [x,y,z] for site/bond ?

    Returns:
        - (str) -- input s if it is valid, otherwise None.

    Note:
        - bond sytles, start:vector, tail;head, vector@center, are accepted.
    """
    if s.count(":"):
        s2 = s.split(":")
    elif s.count(";"):
        s2 = s.split(";")
    elif s.count("@"):
        s2 = s.split("@")
    else:
        return None

    if len(s2) != 2:
        return None
    sa, sb = s2
    sa = validator_site(sa, use_var)
    if sa is None:
        return None
    sb = validator_site(sb, use_var)
    if sb is None:
        return None

    return s


# ==================================================
def validator_site_bond(s, use_var=False):
    """
    Validator for site or bond.

    Args:
        s (str): site or bond string.
        use_var (bool, optional): use [x,y,z] for site/bond ?

    Returns:
        - (str) -- input s if it is valid, otherwise None.

    Note:
        - bond sytles, start:vector, tail;head, vector@center, are accepted.
    """
    if (":" in s) or (";" in s) or ("@" in s):
        return validator_bond(s, use_var)
    else:
        return validator_site(s, use_var)


# ==================================================
def validator_vector_site_bond(s, use_var=False):
    """
    Validator for vector on site or bond.

    Args:
        s (str): vector on site or bond string.
        use_var (bool, optional): use [x,y,z] for site/bond ?

    Returns:
        - (str) -- input s if it is valid, otherwise None.

    Note:
        - bond sytles, start:vector, tail;head, vector@center, are accepted.
    """
    if s.count("#"):
        s2 = s.split("#")
        if len(s2) != 2:
            return None
        v, sb = s2
        v = validator_site(v, use_var)
        if v is None:
            return None
        sb = validator_site_bond(sb, use_var)
        if sb is None:
            return None
    else:
        return None

    return s


# ==================================================
def validator_orbital_site_bond(s, use_var=False):
    """
    Validator for orbital on site or bond.

    Args:
        s (str): orbital on site or bond string.
        use_var (bool, optional): use [x,y,z] for site/bond ?

    Returns:
        - (str) -- input s if it is valid, otherwise None.

    Note:
        - orbital can contain x, y, z, r.
        - bond sytles, start:vector, tail;head, vector@center, are accepted.
    """
    if s.count("#"):
        s2 = s.split("#")
        if len(s2) != 2:
            return None
        o, sb = s2
        o = str_to_numpy(o, None, ["x", "y", "z", "r"], ())
        if o is None:
            return None
        sb = validator_site_bond(sb, use_var)
        if sb is None:
            return None
    else:
        return None

    return s



"""
Utility for axis actor and label, unit cell, and view vector.

This module contains utility for axis, etc.
"""

import vtk
import pyvista as pv
import numpy as np
from math import floor, ceil
from gcoreutils.color_palette import all_colors
from gcoreutils.convert_util import text_to_list
from qtdraw.core.pyvista_widget_setting import widget_detail as detail
from qtdraw.core.pyvista_widget_setting import CHOP, DIGIT


# ==================================================
def _create_label_axes_actor(A, label, size, bold, italic, color, scale):
    """
    Create label only orientation axes actor.

    Args:
        A (numpy.ndarray): (a1, a2, a3) unit vectors, 4x4 [float].
        label (str): axes labels.
        size (int): font size.
        bold (bool): bold face ?
        italic (bool): italic ?
        color (list): axes colors, [[float]], RGB in unit of [0,1].
        scale (float): zoom factor.

    Returns:
        - (pyvista.AxesActor) -- axes actor.
    """
    label = text_to_list(label.replace(" ", ""))

    # font size is zoomed manually.
    size = int(size * scale)

    # set transform for non-orthogonal axes.
    transform = vtk.vtkTransform()
    transform.SetMatrix(A.ravel().tolist())

    # create axes actor with label only.
    lbl = pv.AxesActor()
    lbl.SetShaftTypeToCylinder()
    lbl.SetTipTypeToCone()
    lbl.SetCylinderRadius(0.0)
    lbl.SetConeRadius(0.0)
    lbl.SetUserTransform(transform)

    # set font properties.
    lbl.SetXAxisLabelText(label[0])
    lbl.SetYAxisLabelText(label[1])
    lbl.SetZAxisLabelText(label[2])

    x_p = lbl.GetXAxisCaptionActor2D()
    y_p = lbl.GetYAxisCaptionActor2D()
    z_p = lbl.GetZAxisCaptionActor2D()
    for i, c in zip([x_p, y_p, z_p], color):
        i.GetPositionCoordinate().SetCoordinateSystemToWorld()
        i.GetTextActor().SetTextScaleModeToViewport()
        if bold:
            i.GetCaptionTextProperty().BoldOn()
        else:
            i.GetCaptionTextProperty().BoldOff()
        if italic:
            i.GetCaptionTextProperty().ItalicOn()
        else:
            i.GetCaptionTextProperty().ItalicOff()
        i.GetCaptionTextProperty().SetFontSize(size)
        i.GetCaptionTextProperty().SetColor(*c)

    return lbl


# ==================================================
def _create_axes_actor(
    A,
    label,
    label_size,
    label_bold,
    label_italic,
    label_color,
    scale,
    shaft_color,
    sphere_color,
    shaft_radius,
    tip_radius,
    tip_length,
    tip_resolution,
    sphere_radius,
    theta_phi_resolution,
):
    """
    Create custom axes actor.

    Args:
        A (numpy.ndarray): (a1, a2, a3) unit vectors, 4x4 [float].
        label (str): axes labels.
        label_size (int): font size.
        label_bold (bool): bold face ?
        label_italic (bool): italic ?
        label_color (list): axes label color names, [str].
        scale (float): zoom factor.
        shaft_color (list): axes color names, [str].
        sphere_color (str): center color name.
        shaft_radius (float): axes cylinder radius.
        tip_radius (float): axes tip radius.
        tip_length (float): axes tip length.
        tip_resolution (int): axes tip resolution.
        sphere_radius (float): axes sphere radius.
        theta_phi_resolution (list): axes sphere theta, phi resolution, [int].

    Returns:
        - (vtk.vtkPropAssembly) -- custom axes actor.
    """
    # convert from color name to RGB float.
    shaft_color = [(np.array(all_colors[c][1]) / 255) for c in shaft_color]
    sphere_color = np.array(all_colors[sphere_color][1]) / 255
    # convert from color name to RGB.
    label_color = [all_colors[c][1] for c in label_color]

    # create axes.
    assembly = vtk.vtkPropAssembly()
    for d, c in zip(A[0:3, 0:3].T, shaft_color):
        # axes arrows.
        g = pv.Arrow(
            direction=d,
            shaft_radius=shaft_radius,
            tip_radius=tip_radius,
            tip_length=tip_length,
            tip_resolution=tip_resolution,
        )
        actor = pv.Actor(mapper=pv.DataSetMapper(g))
        actor.GetProperty().SetColor(c)
        assembly.AddPart(actor)

        # dummy axes to keep rotation center as origin.
        g = pv.Sphere(radius=0.0, center=-np.array(d))
        actor = pv.Actor(mapper=pv.DataSetMapper(g))
        assembly.AddPart(actor)

    # center sphere (theta, phi are used differently).
    phi, theta = theta_phi_resolution
    g0 = pv.Sphere(radius=sphere_radius, theta_resolution=theta, phi_resolution=phi)
    actor = pv.Actor(mapper=pv.DataSetMapper(g0))
    actor.GetProperty().SetColor(sphere_color)
    assembly.AddPart(actor)

    # add axes label.
    if label is not None:
        lbl = _create_label_axes_actor(
            A,
            label=label,
            size=label_size,
            bold=label_bold,
            italic=label_italic,
            color=label_color,
            scale=scale,
        )
        assembly.AddPart(lbl)

    return assembly


# ==================================================
def create_axes_widget(
    pv_widget,
    A,
    label="[x,y,z]",
    label_size=28,
    label_bold=True,
    label_italic=False,
    label_color=["black", "black", "black"],
    viewport=True,
):
    """
    Create axes widget.

    Args:
        pv_widget (PyVistaWidget): pyvista widget.
        A (numpy.ndarray): (a1, a2, a3) unit vectors, 4x4 [float].
        label (str, optional): axes labels.
        label_size (int, optional): font size.
        label_bold (bool, optional): bold face ?
        label_italic (bool, optional): italic ?
        label_color (list, optional): axes label color names, [str].
        viewport (bool, optional): set viewport ?

    Note:
        - if label is None, no label is used.
    """
    scale = detail["scale"]
    pickable = detail["pickable"]
    shaft_color = detail["shaft_color"]
    sphere_color = detail["sphere_color"]
    shaft_radius = detail["shaft_radius"]
    tip_radius = detail["tip_radius"]
    tip_length = detail["tip_length"]
    tip_resolution = detail["tip_resolution"]
    sphere_radius = detail["sphere_radius"]
    theta_phi_resolution = detail["theta_phi_resolution"]

    if viewport:
        viewport = detail["viewport"]
    else:
        viewport = [0.0, 0.0, 1.0, 1.0]
        pickable = False

    # create axes actor.
    marker = _create_axes_actor(
        A,
        label=label,
        label_size=label_size,
        label_bold=label_bold,
        label_italic=label_italic,
        label_color=label_color,
        scale=scale,
        shaft_color=shaft_color,
        sphere_color=sphere_color,
        shaft_radius=shaft_radius,
        tip_radius=tip_radius,
        tip_length=tip_length,
        tip_resolution=tip_resolution,
        sphere_radius=sphere_radius,
        theta_phi_resolution=theta_phi_resolution,
    )

    # create axes widget.
    pv_widget.add_orientation_widget(marker, interactive=pickable, viewport=viewport)
    pv_widget.renderer.axes_widget.SetZoom(scale)


# ==================================================
def create_unit_cell(A, origin, lower=None, dimensions=None):
    """
    Create unit cell mesh.

    Args:
        A (numpy.ndarray): (a1, a2, a3) unit vectors, 4x4 [float].
        origin (list or numpy.ndarray): origin, [float].
        lower (list, optional): lower bound indices, [int].
        dimensions (list, optional): repeat times, [int].

    Returns:
        - (pyvista.PolyData) -- unit cel mesh.
    """
    if lower is None:
        lower = [0, 0, 0]
    if dimensions is None:
        dimensions = [1, 1, 1]

    # signle box.
    pts = np.array(
        [
            [0, 0, 0],
            [1, 0, 0],
            [1, 1, 0],
            [0, 1, 0],
            [0, 0, 1],
            [1, 0, 1],
            [1, 1, 1],
            [0, 1, 1],
        ],
        dtype=np.float64,
    )
    shift = np.array([origin] * len(pts), dtype=np.float64)
    pts = pts + shift
    lines = [5, 0, 1, 2, 3, 0, 5, 4, 5, 6, 7, 4, 2, 0, 4, 2, 1, 5, 2, 2, 6, 2, 3, 7]
    box = pv.PolyData(pts, lines=lines)
    box.transform(A, inplace=True)

    # repeated boxes.
    m = pv.ImageData(dimensions=dimensions, origin=lower).cast_to_unstructured_grid()
    m.transform(A, inplace=True)
    p = m.glyph(geom=box, factor=1.0, scale=False, orient=False)

    return p


# ==================================================
def get_view_vector(n, A):
    """
    Get view and viewup.

    Args:
        n (list): view indices of [a1,a2,a3], [int].
        A (numpy.ndarray): (a1, a2, a3) unit vectors, 4x4 [float].

    Returns:
        - (numpy.ndarray) -- view.
        - (numpy.ndarray) -- viewup.
    """
    view = n[0] * A[0:3, 0] + n[1] * A[0:3, 1] + n[2] * A[0:3, 2]
    norm = np.linalg.norm(view)
    view = view / norm

    if np.allclose(view, [0, 1, 0]) or np.allclose(view, [0, 0, -1]):
        viewup = np.array([1, 0, 0])
    elif np.allclose(view, [0, 0, 1]) or np.allclose(view, [-1, 0, 0]):
        viewup = np.array([0, 1, 0])
    elif np.allclose(view, [1, 0, 0]) or np.allclose(view, [0, -1, 0]):
        viewup = np.array([0, 0, 1])
    else:
        vz = np.sqrt(1.0 - view[2] * view[2])
        vx = -view[2] * view[0] / vz
        vy = -view[2] * view[1] / vz
        viewup = np.array([vx, vy, vz], dtype=np.float64)

    return view, viewup


# ==================================================
def create_grid(ilower, dims):
    """
    Create grid point.

    Parameters:
    Args:
        ilower (list): start cell.
        dims (list): range in each dim.

    Returns:
        - (list) -- grid point, [str].
    """
    # range.
    x = np.arange(ilower[0], ilower[0] + dims[0])
    y = np.arange(ilower[1], ilower[1] + dims[1])
    z = np.arange(ilower[2], ilower[2] + dims[2])

    # mesh grid.
    X, Y, Z = np.meshgrid(x, y, z, indexing="ij")

    # transform grid to point.
    grid = np.vstack([X.ravel(), Y.ravel(), Z.ravel()]).T

    grid = [str(i).replace(" ", "") for i in grid.tolist()]

    return grid


# ==================================================
def get_lattice_vector(crystal, cell):
    """
    Get lattice vector.

    Args:
        crystal (str): crystal.
        cell (dict): cell.

    Returns:
        - dict: cell.
        - list: A.
    """
    a = float(cell["a"])
    b = float(cell["b"])
    c = float(cell["c"])
    alpha = float(cell["alpha"])
    beta = float(cell["beta"])
    gamma = float(cell["gamma"])

    if crystal == "monoclinic":
        alpha = 90.0
        gamma = 90.0
    elif crystal == "orthorhombic":
        alpha = 90.0
        beta = 90.0
        gamma = 90.0
    elif crystal in ["trigonal", "hexagonal"]:
        alpha = 90.0
        beta = 90.0
        gamma = 120.0
        b = a
    elif crystal == "tetragonal":
        alpha = 90.0
        beta = 90.0
        gamma = 90.0
        b = a
    elif crystal == "cubic":
        alpha = 90.0
        beta = 90.0
        gamma = 90.0
        b = a
        c = a

    ca = np.cos(alpha * np.pi / 180)
    cb = np.cos(beta * np.pi / 180)
    cc = np.cos(gamma * np.pi / 180)
    sc = np.sin(gamma * np.pi / 180)
    s = 1.0 - ca * ca - cb * cb - cc * cc + 2.0 * ca * cb * cc
    s = max(CHOP, np.sqrt(s))

    a1 = np.array([a, 0, 0]).round(DIGIT).tolist()
    a2 = np.array([b * cc, b * sc, 0]).round(DIGIT).tolist()
    a3 = np.array([c * cb, c * (ca - cb * cc) / sc, c * s / sc]).round(DIGIT).tolist()

    A = np.eye(4)
    A[0:3, 0] = a1
    A[0:3, 1] = a2
    A[0:3, 2] = a3

    A = A.round(DIGIT).tolist()
    cell = {"a": a, "b": b, "c": c, "alpha": alpha, "beta": beta, "gamma": gamma}

    return cell, A


# ==================================================
def get_repeat_range(lower, upper):
    """
    Get repeart range.

    Args:
        lower (list): upper.
        upper (list): lower.

    Returns:
        - list: lower cell.
        - list: size of repeat.
    """
    i1 = [floor(lower[0]), floor(lower[1]), floor(lower[2])]
    i2 = [ceil(upper[0] + CHOP), ceil(upper[1] + CHOP), ceil(upper[2] + CHOP)]
    dims = [i2[0] - i1[0], i2[1] - i1[1], i2[2] - i1[2]]
    for i in range(3):
        dims[i] = max(1, dims[i])

    return i1, dims


# ==================================================
def get_outside_box(point, lower, upper):
    """
    Get indices outside range.

    Args:
        point (numpy.ndarray): a set of points.
        lower (list): lower bound.
        upper (list): upper bound.

    Returns:
        - (numpy.ndarray) -- list of indices.
    """
    xmin, ymin, zmin = lower
    xmax, ymax, zmax = upper

    in_x = (point[:, 0] >= xmin) & (point[:, 0] <= xmax)
    in_y = (point[:, 1] >= ymin) & (point[:, 1] <= ymax)
    in_z = (point[:, 2] >= zmin) & (point[:, 2] <= zmax)

    in_box = in_x & in_y & in_z
    outside = np.where(~in_box)[0]

    return outside



"""
Simple parser for CIF, XSF, and VESTA.

This module contains the parser.
"""

import numpy as np
from qtdraw.parser.util import parse_material, draw_site_bond
from qtdraw.parser.xsf import extract_data_xsf


# ==================================================
def read_draw(filename, widget):
    """
    Read and draw CIF, XSF, VESTA file.

    Args:
        filename (str): filename.
        widget (PyVistaWidget): PyVista widget.

    Returns:
        - (dict) -- all data.
    """
    all_data, site_info, bond_info, symmetrized = parse_material(filename)
    widget.write_info("* " + str(symmetrized))

    name = all_data["status"]["model"]
    draw_site_bond(widget, name, site_info, bond_info)

    if filename.endswith(".xsf"):
        # determine color_range and value.
        data = np.array(extract_data_xsf(filename)["data"])
        d_max = float(data.max())
        d_min = float(data.min())
        d_max = 1.1 * d_max if d_max > 0.0 else 0.9 * d_max
        d_min = 0.9 * d_min if d_min > 0.0 else 1.1 * d_min
        val = 0.5 * (d_max + d_min)
        widget.add_isosurface(data=filename, color="Pastel1", surface="phase", color_range=[d_min, d_max], value=[val])

    return all_data



"""
Info. dialog.

This module provides a dialog for group info. in MultiPie dialog.
"""

import sympy as sp
from PySide6.QtWidgets import QDialog
from gcoreutils.list_util import list_to_table
from gcoreutils.nsarray import NSArray
from multipie.tag.tag_multipole import TagMultipole
from multipie.tag.tag_irrep import TagIrrep
from qtdraw.widget.custom_widget import Layout
from qtdraw.widget.table_view import TableView


# ==================================================
class InfoPanel(QDialog):
    def __init__(self, parent, data, header, title, vertical):
        """
        Info panel.

        Args:
            parent (QWidget): parent.
            data (list): list of latex string without "$", [[str]].
            header (list): header. None is for no header.
            title (str): title of window.
            vertical (bool): show vertical (sequential number) header ?
        """
        super().__init__(parent)
        latex = parent.plugin._pvw._preference["latex"]

        self.setWindowTitle(title)
        self.resize(800, 600)

        layout = Layout(self)
        layout.setContentsMargins(10, 10, 10, 10)

        color = latex["color"]
        size = latex["size"]
        dpi = latex["dpi"]
        table = TableView(self, data, header, vertical, color, size, dpi)
        layout.addWidget(table)

        self.show()


# ==================================================
def show_group_info(
    group,
    name,
    header,
    data,
    vertical,
    parent=None,
):
    """
    Show group info.

    Args:
        group (PointGroup or SpaceGroup): group.
        name (str): title of dialog.
        header (list): header string. None is for no header.
        data (list): list of latex string without "$", [[str]].
        vertical (bool): show vertical (sequential number) header ?
        parent (QWidget, optional): parent.

    Returns:
        - (InfoPanel) -- panel.
    """
    no, _, IS, setting = group.tag.info()
    s = f"No.{no}: {str(group)}, {IS}"
    if setting:
        s += " (" + setting + " setting)"
    title = name + " - " + s

    return InfoPanel(parent, data, header, title, vertical)


# ==================================================
def show_symmetry_operation(group, parent):
    """
    Show symmetry operation panel.

    Args:
        group (PointGroup): point group.
        parent (QWidget): parent.

    Returns:
        - (InfoPanel) -- symmetry operation panel.
    """
    so = group.symmetry_operation

    name = "symmetry_operation"
    header = ["symbol", "polar vector", "axial vector", "det"]

    data = []
    if not group.tag.is_point_group():
        ps = ["+" + i for i in so.plus_set.latex()]
        data.append(ps + [""] * (4 - len(ps)))
        data.append(["", "", "", ""])

    pso = so.mat(axial=False)
    aso = so.mat(axial=True)
    for no in range(len(so)):
        p = pso[no][0:3, :]
        a = aso[no][0:3, :]
        data.append([so.full[no].latex(), p.latex(), a.latex(), p[0:3, 0:3].det().latex()])

    return show_group_info(group, name, header, data, False, parent)


# ==================================================
def show_character_table(group, parent):
    """
    Show character table panel.

    Args:
        group (PointGroup): point group.
        parent (QWidget): parent.

    Returns:
        - (InfoPanel) -- character table panel.
    """
    ch = group.character

    name = "character"
    n = len(ch.symmetry_operation())
    header = ["irrep."] + [f"SO{i+1}" for i in range(n)]

    data = []
    data.append([""] + [i.latex() for i in ch.symmetry_operation()])
    for r in ch.irrep_list:
        data.append([r.latex()] + [sp.latex(j) for j in ch.character(r)])

    return show_group_info(group, name, header, data, False, parent)


# ==================================================
def show_wyckoff(group, parent):
    """
    Show Wyckoff position panel.

    Args:
        group (PointGroup): point group.
        parent (QWidget): parent.

    Returns:
        - (InfoPanel) -- Wyckoff position panel.
    """
    wp = group.wyckoff
    so = group.symmetry_operation

    name = "wyckoff_position"
    n = list(wp.keys())[-1].n
    if not group.tag.is_point_group():
        ns = len(so.plus_set)
        n = n // ns
    n = min(10, n)
    header = ["WP"] + [str(i + 1) for i in range(n)]

    data = []
    if not group.tag.is_point_group():
        ps = ["+" + i for i in so.plus_set.latex()]
        data.append([""] + ps + [""] * (n - len(ps)))
        data.append([""] * (n - len(ps) + 2))

    for w in wp.keys():
        row = list_to_table(wp.position(w).latex(), n, "")
        for i, r in enumerate(row):
            if i == 0:
                data.append([str(w)] + r)
            else:
                data.append([""] + r)

    return show_group_info(group, name, header, data, False, parent)


# ==================================================
def show_product_table(group, parent):
    """
    Show product table panel.

    Args:
        group (PointGroup): point group.
        parent (QWidget): parent.

    Returns:
        - (InfoPanel) -- product table panel.
    """
    so = group.symmetry_operation

    name = "product_table"
    op = [i.latex() for i in so.keys()]
    header = ["SO"] + [f"SO{i+1}" for i in range(len(op))]

    data = []
    data.append([""] + op)
    for i in so.keys():
        data.append([i.latex()] + [so.product(i, j).latex() for j in so.keys()])

    return show_group_info(group, name, header, data, False, parent)


# ==================================================
def show_harmonics(group, rank, head, parent):
    """
    Show harmonics panel.

    Args:
        group (PointGroup): point group.
        rank (int): harmonics rank.
        head (str): harmonics type.
        parent (QWidget): parent.

    Returns:
        - (InfoPanel) -- harmonics panel.
    """
    pgh = group.harmonics
    head = head.replace("T", "Q").replace("M", "G")
    hs = pgh.select(rank=rank, head=head)

    name = head + "_harmonics"
    header = ["symbol", "rank", "irrep.", "mul.", "comp.", "expression", "definition"]

    data = []
    for h in hs:
        mul = h.tag.mul
        comp = h.tag.comp
        irrep = h.tag.tag_irrep()
        dim = irrep.dim
        if mul < 1:
            mul = ""
        if dim == 1:
            comp = ""
        data.append(
            [
                h.latex(),
                rank,
                irrep.latex(),
                mul,
                comp,
                h.expression(v=NSArray.vector3d(head)).latex(),
                h.definition().latex(),
            ]
        )

    return show_group_info(group, name, header, data, True, parent)


# ==================================================
def show_harmonics_decomp(group, rank, head, to_pg, parent):
    """
    Show harmonics decomposition panel.

    Args:
        group (PointGroup): point group.
        rank (int): harmonics rank.
        head (str): harmonics type.
        to_pg (str): destination point group.
        parent (QWidget): parent.

    Returns:
        - (InfoPanel) -- harmonics decomposition panel.
    """
    pgh = group.harmonics

    if to_pg.count(" ") > 0:
        to_pg = to_pg.split(" ")[1]

    head = head.replace("T", "Q").replace("M", "G")
    axial = head == "G"
    hs = pgh.select(rank=rank, head=head)
    decomp = group.irrep_decomposition(to_pg, rank, axial)

    name = head + "_harmonics decomposition to " + to_pg
    header = [str(group), " \u21d2 " + to_pg]

    data = []
    for h, d in zip(hs, decomp):
        ex = sp.S(0)
        for c, b in d[1]:
            ex += c * TagMultipole(b).symbol()
        data.append([h.latex(), sp.latex(ex)])

    return show_group_info(group, name, header, data, True, parent)


# ==================================================
def show_virtual_cluster(group, wp, parent):
    """
    Show virtual cluster panel.

    Args:
        group (PointGroup): point group.
        wp (str): Wyckoff position.
        parent (QWidget): parent.

    Returns:
        - (InfoPanel) -- virtual cluster panel.
    """
    basis, site = group.virtual_cluster_basis(wyckoff=wp)
    n = min(10, len(site))

    name = f"virtual cluster ({wp})"
    header = ["symbol"] + [f"S{i+1}" for i in range(n)]

    data = []
    row = list_to_table(site.latex(), n, "")
    for i, r in enumerate(row):
        if i == 0:
            data.append([r"\text{site (r.c.)}"] + r)
        else:
            data.append([""] + r)
    for tag, v in basis.items():
        row = list_to_table(NSArray(v.tolist(), "scalar").latex(), n, "")
        for i, r in enumerate(row):
            if i == 0:
                t = tag.latex()
            else:
                t = ""
            data.append([t] + r)

    return show_group_info(group, name, header, data, False, parent)


# ==================================================
def show_atomic_multipole(group, bra, ket, am, parent):
    """
    Show atomic multipole panel.

    Args:
        group (PointGroup): point group.
        bra (list): bra basis list.
        ket (list): ket basis list.
        am (dict): atomic multipoles.
        parent (QWidget): parent.

    Returns:
        - (InfoPanel) -- symmetry operation panel.
    """
    name = "atomic multipole"
    header = ["symbol", "rank", "irrep.", "s", "k", "basis"]

    data = []
    s = r"\langle " + ",".join(bra) + r"|" + ",".join(ket) + r"\rangle"
    data.append([""] * 5 + [s])
    data.append([""] * 6)
    for tag, am1 in am.items():
        data.append(
            [
                tag.latex(),
                tag.rank,
                TagIrrep(tag.irrep).latex(),
                tag.s,
                tag.k,
                am1.latex(),
            ]
        )

    return show_group_info(group, name, header, data, False, parent)


# ==================================================
def show_response(group, rank, r_type, parent):
    """
    Show response tensor panel.

    Args:
        group (PointGroup): point group.
        rank (int): response tensor rank.
        r_type (str): response tensor type.
        parent (QWidget): parent.

    Returns:
        - (InfoPanel) -- response tensor panel.
    """
    dic = {"Q": "E_polar", "T": "M_polar", "G": "E_axial", "M": "M_axial"}
    pgr = group.response
    head = r_type
    rt = pgr.select(rank=rank, head=head)

    name = dic[head] + "_response"
    header = ["symbol", "tensor"]

    data = []
    for tag in rt.keys():
        M = rt[tag]
        data.append([tag.latex(), M.latex()])

    data.append(["", ""])
    data.append([r"\text{component}", r"\text{multipole}"])
    for tag in rt.keys():
        def_mul = rt.definition[tag]
        for t, d in def_mul.items():
            data.append([sp.latex(t), sp.latex(d)])

    return show_group_info(group, name, header, data, False, parent)



"""
Simple PDF viewer.

This module contains a class for simple PDF viewer.
It can load PDF file, and save the same file to
different directory (just by copying it).
"""

from PySide6.QtWidgets import QFileDialog, QGridLayout, QWidget, QPushButton
from PySide6.QtPdf import QPdfDocument
from PySide6.QtPdfWidgets import QPdfView
from shutil import copy
import os


# ==================================================
class PDFViewer(QWidget):

    # ==================================================
    def __init__(self, filename=None):
        """
        Simple PDF viewer.

        Args:
            filename (str, optional): file name to load.
        """
        super().__init__()

        self.document = QPdfDocument(self)
        if filename is not None:
            self.document.load(filename)
            self.filename = filename

        view = QPdfView(self)
        view.setDocument(self.document)
        view.setPageMode(QPdfView.PageMode.MultiPage)
        view.setZoomMode(QPdfView.ZoomMode.FitToWidth)

        open_btn = QPushButton("Open")
        open_btn.clicked.connect(self.load)
        save_btn = QPushButton("Save")
        save_btn.clicked.connect(self.save)

        layout = QGridLayout(self)
        layout.setContentsMargins(0, 0, 0, 0)
        layout.setSpacing(7)
        layout.addWidget(open_btn, 0, 0, 1, 1)
        layout.addWidget(save_btn, 0, 1, 1, 1)
        layout.addWidget(view, 1, 0, 1, 12)

        self.setLayout(layout)
        self.resize(800, 1000)
        self.set_title()

    # ==================================================
    def load(self):
        """
        Load PDF file.
        """
        self.filename, _ = QFileDialog.getOpenFileName(self, filter="PDF (*.pdf)")
        if self.filename != "":
            self.document.load(self.filename)
            self.set_title(self.filename)

    # ==================================================
    def save(self):
        """
        Save PDF file (default = current directory).
        """
        full = os.path.join(os.getcwd(), os.path.split(self.filename)[1])
        filename, _ = QFileDialog.getSaveFileName(self, dir=full, filter="PDF (*.pdf)")
        if filename != "":
            copy(self.filename, filename)

    # ==================================================
    def set_title(self, filename=None):
        """
        Set window title.

        Args:
            filename (str, optional): file name.
        """
        if filename is None:
            filename = "None"

        self.setWindowTitle(f"PDF Viewer - {filename}")
